@@ -28,6 +28,9 @@ import os.path
 import re
 from . import _debugprint
 
+__all__ = ['ppdMakeModelSplit',
+           'PPDs']
+
 def ppdMakeModelSplit (ppd_make_and_model):
     """
     Split a ppd-make-and-model string into a canonical make and model pair.
@@ -82,7 +85,7 @@ def ppdMakeModelSplit (ppd_make_and_model):
         model = ppd_make_and_model
     elif re.search ("^(okipage|microline)", \
                       ppd_make_and_model, re.I):
-        make = "Okidata"
+        make = "Oki"
         model = ppd_make_and_model
     elif re.search ("^(konica[\s_-]*minolta)", \
                       ppd_make_and_model, re.I):
@@ -138,6 +141,7 @@ def ppdMakeModelSplit (ppd_make_and_model):
         model = model[:wth]
 
     make = re.sub (r"(?i)KONICA[\s_-]*MINOLTA", "KONICA MINOLTA", make, 1)
+    make = re.sub (r"(?i)HEWLETT[\s_-]PACKARD", "HP", make, 1)
     model = re.sub (r"(?i)\s*\(recommended\)", "", model)
     model = re.sub (r"(?i)\s*-\s*PostScript\b", "", model)
     model = re.sub (r"(?i)\s*\bseries\b", "", model)
@@ -147,9 +151,11 @@ def ppdMakeModelSplit (ppd_make_and_model):
     model = re.sub (r"(?i)\s*\(Bluetooth\)", "", model)
     model = re.sub (r"(?i)\s*-\s*(RC|Ver(|sion))\s*-*\s*[0-9\.]+", "", model)
     model = re.sub (r"(?i)\s*-\s*(RC|Ver(|sion))\b", "", model)
-    model = re.sub (r"(?i)\s*-\s*$", "", model)
+    model = re.sub (r"(?i)\s*PostScript\s*$", "", model)
+    model = re.sub (r"(?i)\s*\(\s*\)", "", model)
+    model = re.sub (r"(?i)\s*[\-\/]\s*$", "", model)
 
-    for mfr in [ "Apple", "Canon", "Epson", "Lexmark", "Okidata" ]:
+    for mfr in [ "Apple", "Canon", "Epson", "Lexmark", "Oki" ]:
         if make == mfr.upper ():
             make = mfr
 
@@ -158,6 +164,7 @@ def ppdMakeModelSplit (ppd_make_and_model):
 
 # Some drivers are just generally better than others.
 # Here is the preference list:
+DRIVER_TYPE_DOWNLOADED_NOW = 5
 DRIVER_TYPE_FOOMATIC_RECOMMENDED_NON_POSTSCRIPT = 8
 DRIVER_TYPE_VENDOR = 10
 DRIVER_TYPE_FOOMATIC_RECOMMENDED_POSTSCRIPT = 15
@@ -175,12 +182,13 @@ DRIVER_TYPE_FOOMATIC_GENERIC = 90
 DRIVER_DOES_NOT_WORK = 999
 def _getDriverType (ppdname, ppds=None):
     """Decides which of the above types ppdname is."""
-    if ppdname.startswith ("gutenprint"):
-        if ppdname.find ("/simple/") != -1:
+    if ppdname.find ("gutenprint") != -1:
+        if (ppdname.find ("/simple/") != -1 or
+            ppdname.find (".sim-") != -1):
             return DRIVER_TYPE_GUTENPRINT_NATIVE_SIMPLIFIED
         else:
             return DRIVER_TYPE_GUTENPRINT_NATIVE
-    if ppdname.find ("SpliX")!= -1:
+    if ppdname.find ("splix")!= -1:
         return DRIVER_TYPE_SPLIX
     if (ppdname.find (":") == -1 and
         ppdname.find ("/cups-included/") != -1):
@@ -238,7 +246,6 @@ class PPDs:
 	@param language: language name, as given by the first element
         of the pair returned by locale.getlocale()
         """
-        print _debugprint
         self.ppds = ppds.copy ()
         self.makes = None
         self.ids = None
@@ -311,7 +318,7 @@ class PPDs:
             models_list = self.makes[make].keys ()
         except KeyError:
             return []
-        models_list.sort (cups.modelSort)
+        models_list.sort (key=unicode.lower, cmp=cups.modelSort)
         return models_list
 
     def getInfoFromModel (self, make, model):
@@ -320,7 +327,7 @@ class PPDs:
         particular printer model, given its make and model name.
 
 	@returns: a dict, indexed by ppd-name, of dicts representing
-        PPDs (as given by cups.Connection.getPPDs)"
+        PPDs (as given by cups.Connection.getPPDs)
 	"""
         self._init_makes ()
         try:
@@ -335,7 +342,8 @@ class PPDs:
 	"""
         return self.ppds[ppdname]
 
-    def orderPPDNamesByPreference (self, ppdnamelist=[]):
+    def orderPPDNamesByPreference (self, ppdnamelist=[],
+                                   downloadedfiles=[]):
         """
 
 	Sort a list of PPD names by (hard-coded) preferred driver
@@ -353,6 +361,11 @@ class PPDs:
         mfg, mdl = ppdMakeModelSplit (make_model)
         def getDriverTypeWithBias (x, mfg):
             t = _getDriverType (x, ppds=self)
+            for file in downloadedfiles:
+                (path, slash, filename) = file.rpartition ("/")
+                (xpath, xslash, xfilename) = x.rpartition ("/")
+                if filename == xfilename:
+                    return DRIVER_TYPE_DOWNLOADED_NOW
             if mfg == "HP":
                 if t == DRIVER_TYPE_FOOMATIC_HPIJS:
                     # Prefer HPIJS for HP devices.
@@ -374,29 +387,34 @@ class PPDs:
                 else:
                     return -1
 
-            # Prefer C locale localized PPDs to other languages,
-            # just because we don't know the user's locale.
+            # Prefer C locale localized PPDs to other languages, just
+            # because we don't know the user's locale.  This only
+            # applies to PPDs provided by gutenprint 5.0; in 5.2 the
+            # PPDs are localized properly.
             def is_C_locale (x):
-                while x:
-                    i = x.find ("C")
-                    if i == -1:
-                        return False
-                    lword = False
-                    if i == 0:
-                        lword = True
-                    elif x[i - 1] not in string.letters:
-                        lword = True
+                try:
+                    while x:
+                        i = x.find ("C")
+                        if i == -1:
+                            return False
+                        lword = False
+                        if i == 0:
+                            lword = True
+                        elif x[i - 1] not in string.letters:
+                            lword = True
 
-                    if lword:
-                        rword = False
-                        if i == (len (x) - 1):
-                            rword = True
-                        elif x[i + 1] not in string.letters:
-                            rword = True
-                        if rword:
-                            return True
+                        if lword:
+                            rword = False
+                            if i == (len (x) - 1):
+                                rword = True
+                            elif x[i + 1] not in string.letters:
+                                rword = True
+                            if rword:
+                                return True
                         
-                    x = x[i + 1:]
+                        x = x[i + 1:]
+                except UnicodeDecodeError:
+                    return False
 
             ca = is_C_locale (a)
             cb = is_C_locale (b)
@@ -424,7 +442,8 @@ class PPDs:
         return ppdnamelist
 
     def getPPDNameFromDeviceID (self, mfg, mdl, description="",
-                                commandsets=[], uri=None):
+                                commandsets=[], uri=None,
+                                downloadedfiles=[]):
         """
 	Obtain a best-effort PPD match for an IEEE 1284 Device ID.
 	The status is one of:
@@ -487,6 +506,8 @@ class PPDs:
             # Try again with replacements.
             if mfgl == "hewlett-packard":
                 mfgl = "hp"
+            elif mfgl == "lexmark international":
+                mfgl = "lexmark"
 
         # Remove manufacturer name from model field
         ppdnamelist2 = None
@@ -573,11 +594,14 @@ class PPDs:
         # Now we have to choose the "best" one.  This is quite tricky
         # to decide, so let's sort them in order of preference and
         # take the first.
-        ppdnamelist = self.orderPPDNamesByPreference (ppdnamelist)
+        ppdnamelist = self.orderPPDNamesByPreference (ppdnamelist,
+                                                      downloadedfiles)
         _debugprint (str (ppdnamelist))
 
         if not id_matched:
-            print "No ID match for device %s:" % uri
+            sanitised_uri = re.sub (pattern="//[^@]*@/?", repl="//",
+                                    string=str (uri))
+            print "No ID match for device %s:" % sanitised_uri
             print "  <manufacturer>%s</manufacturer>" % mfg
             print "  <model>%s</model>" % mdl
             print "  <description>%s</description>" % description
@@ -670,6 +694,8 @@ class PPDs:
                         digits_start = i
                     digits_end = i
                     digits += 1
+                elif digits_start != -1:
+                    break
             digits_end += 1
             modelnumber = 0
             if digits > 0:
@@ -821,10 +847,10 @@ class PPDs:
 def _show_help():
     print "usage: ppds.py [--deviceid] [--list-models] [--list-ids] [--debug]"
 
-def self_test():
+def _self_test(argv):
     import sys, getopt
     try:
-        opts, args = getopt.gnu_getopt (sys.argv[1:], '',
+        opts, args = getopt.gnu_getopt (argv[1:], '',
                                         ['help',
                                          'deviceid',
                                          'list-models',
@@ -888,6 +914,10 @@ def self_test():
     print "\nID matching tests\n"
 
     idlist = [
+        # Format is:
+        # (ID string, max status code, expected driver RE match)
+
+        # Specific models
         ("MFG:EPSON;CMD:ESCPL2,BDC,D4,D4PX;MDL:Stylus D78;CLS:PRINTER;"
          "DES:EPSON Stylus D78;", 1, 'Epson Stylus D68'),
         ("MFG:Hewlett-Packard;MDL:LaserJet 1200 Series;"
@@ -907,6 +937,11 @@ def self_test():
          "DES:Canon iP3000;VER:1.09;STA:10;FSI:03;", 1, "Canon PIXMA iP3000"),
         ("MFG:HP;MDL:Deskjet 5400 series;CMD:MLC,PCL,PML,DW-PCL,DESKJET,DYN;"
          "1284.4DL:4d,4e,1;CLS:PRINTER;DES:5440;", 1, "HP DeskJet 5440"),
+        ("MFG:Hewlett-Packard;MDL:HP LaserJet 3390;"
+         "CMD:PJL,MLC,PCL,POSTSCRIPT,PCLXL;",
+         0, "HP LaserJet 3390.*Postscript"),
+
+        # Generic models
         ("MFG:New;MDL:Unknown PS Printer;CMD:POSTSCRIPT;",
          2, "Generic postscript printer"),
         ("MFG:New;MDL:Unknown PCL6 Printer;CMD:PCLXL;", 2, "Generic PCL 6"),
@@ -919,7 +954,7 @@ def self_test():
         ]
 
     if stdin_deviceid:
-        idlist = [raw_input ('Device ID: ')]
+        idlist = [(raw_input ('Device ID: '), 2, '')]
 
     all_passed = True
     for id, max_status_code, modelre in idlist:
