@@ -22,20 +22,23 @@ import gtk
 from debug import *
 
 _ = lambda x: x
+N_ = lambda x: x
 def set_gettext_function (fn):
     global _
     _ = fn
 
 class AuthDialog(gtk.Dialog):
-    AUTH_FIELD={'username': _("Username:"),
-                'password': _("Password:"),
-                'domain': _("Domain:")}
+    AUTH_FIELD={'username': N_("Username:"),
+                'password': N_("Password:"),
+                'domain': N_("Domain:")}
 
-    def __init__ (self, title=_("Authentication"), parent=None,
+    def __init__ (self, title=None, parent=None,
                   flags=gtk.DIALOG_MODAL | gtk.DIALOG_NO_SEPARATOR,
                   buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                            gtk.STOCK_OK, gtk.RESPONSE_OK),
                   auth_info_required=['username', 'password']):
+        if title == None:
+            title = _("Authentication")
         gtk.Dialog.__init__ (self, title, parent, flags, buttons)
         self.auth_info_required = auth_info_required
         self.set_default_response (gtk.RESPONSE_OK)
@@ -60,7 +63,7 @@ class AuthDialog(gtk.Dialog):
         self.field_entry = []
         for i in range (num_fields):
             field = auth_info_required[i]
-            label = gtk.Label (self.AUTH_FIELD.get (field, field))
+            label = gtk.Label (_(self.AUTH_FIELD.get (field, field)))
             label.set_alignment (0, 0.5)
             table.attach (label, 0, 1, i, i + 1)
             entry = gtk.Entry ()
@@ -93,10 +96,22 @@ class AuthDialog(gtk.Dialog):
         self.field_entry[i].grab_focus ()
 
 class Connection:
-    def __init__ (self, parent=None, try_as_root=True):
+    def __init__ (self, parent=None, try_as_root=True,
+                  host=None, port=None, encryption=None):
+        if host != None:
+            cups.setServer (host)
+        if port != None:
+            cups.setPort (port)
+        if encryption != None:
+            cups.setEncryption (encryption)
+
         self._use_password = ''
         self._parent = parent
         self._try_as_root = try_as_root
+        self._use_user = cups.getUser ()
+        self._server = cups.getServer ()
+        self._port = cups.getPort()
+        self._encryption = cups.getEncryption ()
         self._connect ()
         self._prompt_allowed = True
 
@@ -107,9 +122,18 @@ class Connection:
         self._prompt_allowed = allowed
 
     def _connect (self):
-        self._connection = cups.Connection ()
-        self._use_user = cups.getUser ()
-        self._server = cups.getServer ()
+        cups.setUser (self._use_user)
+        try:
+            self._connection = cups.Connection (host=self._server,
+                                                port=self._port,
+                                                encryption=self._encryption)
+        except TypeError:
+            # Parameters for Connection() require pycups >= 1.9.40.
+            cups.setServer (self._server)
+            cups.setPort (self._port)
+            cups.setEncryption (self._encryption)
+            self._connection = cups.Connection ()
+
         self._user = self._use_user
         debugprint ("Connected as user %s" % self._user)
         methodtype = type (self._connection.getPrinters)
@@ -126,7 +150,15 @@ class Connection:
 
     def _authloop (self, fname, fn, *args, **kwds):
         self._passes = 0
+        c = self._connection
         while self._perform_authentication () != 0:
+            if c != self._connection:
+                # We have reconnected.
+                fn = getattr (self._connection, fname)
+                c = self._connection
+
+            cups.setUser (self._use_user)
+
             try:
                 result = fn.__call__ (*args, **kwds)
 
@@ -137,20 +169,26 @@ class Connection:
                         raise cups.IPPError (cups.IPP_NOT_AUTHORIZED, '')
                 break
             except cups.IPPError, (e, m):
-                if not self._cancel and e == cups.IPP_NOT_AUTHORIZED:
-                    self._failed ()
+                if not self._cancel and (e == cups.IPP_NOT_AUTHORIZED or
+                                         e == cups.IPP_FORBIDDEN):
+                    self._failed (e == cups.IPP_FORBIDDEN)
                 else:
+                    if self._cancel and not self._cannot_auth:
+                        raise cups.IPPError (0, _("Operation canceled"))
+
                     raise
             except cups.HTTPError, (s,):
-                if not self._cancel and s == cups.HTTP_UNAUTHORIZED:
-                    self._failed ()
+                if not self._cancel and (s == cups.HTTP_UNAUTHORIZED or
+                                         s == cups.HTTP_FORBIDDEN):
+                    self._failed (s == cups.HTTP_FORBIDDEN)
                 else:
                     raise
 
         return result
 
-    def _failed (self):
+    def _failed (self, forbidden=False):
         self._has_failed = True
+        self._forbidden = forbidden
 
     def _password_callback (self, prompt):
         debugprint ("Got password callback")
@@ -169,12 +207,16 @@ class Connection:
             # Haven't yet tried the operation.  Set the password
             # callback and return > 0 so we try it for the first time.
             self._has_failed = False
+            self._forbidden = False
             self._auth_called = False
             self._cancel = False
+            self._cannot_auth = False
+            self._dialog_shown = False
             cups.setPasswordCB (self._password_callback)
             debugprint ("Authentication: password callback set")
             return 1
 
+        debugprint ("Forbidden: %s" % self._forbidden)
         if not self._has_failed:
             # Tried the operation and it worked.  Return 0 to signal to
             # break out of the loop.
@@ -184,20 +226,21 @@ class Connection:
         # Reset failure flag.
         self._has_failed = False
 
-        if self._passes == 2:
+        if self._passes >= 2:
             # Tried the operation without a password and it failed.
             if (self._try_as_root and
                 self._user != 'root' and
-                self._server[0] == '/'):
+                (self._server[0] == '/' or self._forbidden)):
                 # This is a UNIX domain socket connection so we should
-                # not have needed a password, and so the operation must
-                # not be something that the current user is authorised to
-                # do.  They need to try as root, and supply the password.
-                # However, to get the right prompt, we need to try as
-                # root but with no password first.
+                # not have needed a password (or it is not a UDS but
+                # we got an HTTP_FORBIDDEN response), and so the
+                # operation must not be something that the current
+                # user is authorised to do.  They need to try as root,
+                # and supply the password.  However, to get the right
+                # prompt, we need to try as root but with no password
+                # first.
                 debugprint ("Authentication: Try as root")
                 self._use_user = 'root'
-                cups.setUser (self._use_user)
                 self._auth_called = False
                 self._connect ()
                 return 1
@@ -211,25 +254,48 @@ class Connection:
             # We aren't even getting a chance to supply credentials.
             debugprint ("Authentication: giving up")
             self._cancel = True
+            self._cannot_auth = True
             return 1
 
         # Reset the flag indicating whether we were given an auth callback.
         self._auth_called = False
 
+        # If we're previously prompted, explain why we're prompting again.
+        if self._dialog_shown:
+            d = gtk.MessageDialog (self._parent,
+                                   gtk.DIALOG_MODAL |
+                                   gtk.DIALOG_DESTROY_WITH_PARENT,
+                                   gtk.MESSAGE_ERROR,
+                                   gtk.BUTTONS_CLOSE)
+            d.set_title (_("Not authorized"))
+            d.set_markup ('<span weight="bold" size="larger">' +
+                          _("Not authorized") + '</span>\n\n' +
+                          _("The password may be incorrect."))
+            d.run ()
+            d.destroy ()
+
         # Prompt.
         d = AuthDialog (parent=self._parent)
         d.set_prompt (self._prompt)
         d.set_auth_info ([self._use_user, ''])
         d.field_grab_focus ('password')
+        d.set_keep_above (True)
+        d.show_all ()
+        d.show_now ()
+        gtk.gdk.keyboard_grab (d.window, True)
+        gtk.gdk.pointer_grab (d.window, True)
+        self._dialog_shown = True
         response = d.run ()
-        d.hide ()
-
-        if response == gtk.RESPONSE_CANCEL:
-            self._cancel = True
-            return -1
-
+        gtk.gdk.pointer_ungrab ()
+        gtk.gdk.keyboard_ungrab ()
         (self._use_user,
          self._use_password) = d.get_auth_info ()
+        d.destroy ()
+
+        if (response == gtk.RESPONSE_CANCEL or
+            response == gtk.RESPONSE_DELETE_EVENT):
+            self._cancel = True
+            return -1
 
         cups.setUser (self._use_user)
         debugprint ("Authentication: Reconnect")
