@@ -31,6 +31,7 @@ import gtk.glade
 import monitor
 import os
 import pango
+import pwd
 import sys
 import time
 
@@ -48,7 +49,7 @@ from statereason import StateReason
 statereason.set_gettext_function (_)
 errordialogs.set_gettext_function (_)
 
-pkgdata = config.Paths ().get_path ('pkgdatadir')
+pkgdata = config.pkgdatadir
 GLADE="applet.glade"
 ICON="printer"
 SEARCHING_ICON="document-print-preview"
@@ -139,6 +140,14 @@ class PrinterURIIndex:
 
 
 class JobViewer (monitor.Watcher):
+    required_job_attributes = set(['job-hold-until',
+                                   'job-k-octets',
+                                   'job-name',
+                                   'job-originating-user-name',
+                                   'job-printer-uri',
+                                   'job-state',
+                                   'time-at-creation'])
+
     def __init__(self, bus=None, loop=None, service_running=False,
                  trayicon=False, suppress_icon_hide=False,
                  my_jobs=True, specific_dests=None, exit_handler=None,
@@ -155,16 +164,16 @@ class JobViewer (monitor.Watcher):
         self.jobiters = {}
         self.active_jobs = set() # of job IDs
         self.stopped_job_prompts = set() # of job IDs
-        self.which_jobs = "not-completed"
         self.printer_state_reasons = {}
         self.num_jobs_when_hidden = 0
         self.connecting_to_device = {} # dict of printer->time first seen
         self.state_reason_notifications = {}
+        self.auth_notifications = {} # by job ID
+        self.auth_info_dialogs = {} # by job ID
         self.job_creation_times_timer = None
         self.special_status_icon = False
         self.new_printer_notifications = {}
         self.reasoniters = {}
-        self.auth_info_dialog = None
 
         glade_dir = os.environ.get ("SYSTEM_CONFIG_PRINTER_GLADE",
                                     pkgdata)
@@ -287,17 +296,36 @@ class JobViewer (monitor.Watcher):
         if bus == None:
             bus = dbus.SystemBus ()
 
+        self.set_process_pending (True)
+        self.host = cups.getServer ()
+        self.port = cups.getPort ()
+        self.encryption = cups.getEncryption ()
         self.monitor = monitor.Monitor (self, bus=bus, my_jobs=my_jobs,
-                                        specific_dests=specific_dests)
+                                        specific_dests=specific_dests,
+                                        host=self.host, port=self.port,
+                                        encryption=self.encryption)
 
         if not self.trayicon:
             self.MainWindow.show ()
 
     def cleanup (self):
         self.monitor.cleanup ()
+
+        # Close any open notifications.
+        for l in [self.new_printer_notifications.values (),
+                  self.auth_notifications.values (),
+                  self.state_reason_notifications.values ()]:
+            for notification in l:
+                if notification.get_data ('closed') != True:
+                    notification.close ()
+                    notification.set_data ('closed', True)
+
         if self.exit_handler:
             self.exit_handler (self)
 
+    def set_process_pending (self, whether):
+        self.process_pending_events = whether
+
     # Handle "special" status icon
     def set_special_statusicon (self, iconname):
         self.special_status_icon = True
@@ -307,19 +335,17 @@ class JobViewer (monitor.Watcher):
     def unset_special_statusicon (self):
         self.special_status_icon = False
         self.statusicon.set_from_pixbuf (self.saved_statusicon_pixbuf)
+        self.set_statusicon_visibility ()
 
     def notify_new_printer (self, printer, notification):
         self.new_printer_notifications[printer] = notification
         notification.set_data ('printer-name', printer)
         notification.connect ('closed', self.on_new_printer_notification_closed)
         self.set_statusicon_visibility ()
-        # Let the icon show itself, ready for the notification
-        while gtk.events_pending ():
-            gtk.main_iteration ()
         notification.attach_to_status_icon (self.statusicon)
         notification.show ()
 
-    def on_new_printer_notification_closed (self, notification):
+    def on_new_printer_notification_closed (self, notification, reason=None):
         printer = notification.get_data ('printer-name')
         del self.new_printer_notifications[printer]
         self.set_statusicon_visibility ()
@@ -366,10 +392,10 @@ class JobViewer (monitor.Watcher):
 
     def on_show_completed_jobs_activate(self, menuitem):
         if menuitem.get_active():
-            self.monitor.which_jobs = "all"
+            which_jobs = "all"
         else:
-            self.monitor.which_jobs = "not-completed"
-        self.monitor.refresh()
+            which_jobs = "not-completed"
+        self.monitor.refresh(which_jobs=which_jobs, refresh_all=False)
 
     def on_show_printer_status_activate(self, menuitem):
         if self.show_printer_status.get_active ():
@@ -445,18 +471,58 @@ class JobViewer (monitor.Watcher):
     def on_troubleshoot_quit(self, troubleshooter):
         del self.troubleshooter
 
-    def add_job (self, job, data):
+    def add_job (self, job, data, connection=None):
         store = self.store
         iter = self.store.append (None)
         store.set_value (iter, 0, job)
         store.set_value (iter, 1, data.get('job-originating-user-name',
                                            _("Unknown")))
         store.set_value (iter, 2, data.get('job-name', _("Unknown")))
+        debugprint ("Job %d added" % job)
         self.jobiters[job] = iter
-        self.update_job (job, data)
-        self.update_job_creation_times ()
+        if not self.job_creation_times_timer:
+            t = gobject.timeout_add (1000, self.update_job_creation_times)
+            self.job_creation_times_timer = t
+
+        self.update_job (job, data, connection=connection)
+
+    def update_job (self, job, data, connection=None):
+        # Fetch required attributes for this job if they are missing.
+        r = self.required_job_attributes - set (data.keys ())
+        if (data.get ('job-state', cups.IPP_JOB_HELD) != cups.IPP_JOB_HELD):
+            r -= set (['job-hold-until'])
+
+        if r:
+            attrs = None
+            try:
+                if connection == None:
+                    try:
+                        connection = cups.Connection (host=self.host,
+                                                      port=self.port,
+                                                      encryption=self.encryption)
+                    except TypeError:
+                        # Requires pycups >= 1.9.40.
+                        cups.setServer (self.host)
+                        cups.setPort (self.port)
+                        cups.setEncryption (self.encryption)
+                        connection = cups.Connection ()
+
+                debugprint ("requesting %s" % r)
+                r = list (r)
+                try:
+                    attrs = connection.getJobAttributes (job,
+                                                         requested_attributes=r)
+                except TypeError:
+                    # requested_attributes requires pycups >= 1.9.42
+                    attrs = connection.getJobAttributes (jobid)
+            except RuntimeError:
+                pass
+            except AttributeError:
+                pass
+
+            if attrs:
+                data.update (attrs)
 
-    def update_job (self, job, data):
         store = self.store
         iter = self.jobiters[job]
         self.jobs[job] = data
@@ -470,17 +536,24 @@ class JobViewer (monitor.Watcher):
         store.set_value (iter, 4, size)
 
         state = None
+        job_requires_auth = False
         if data.has_key ('job-state'):
             try:
                 jstate = data['job-state']
                 s = int (jstate)
-                state = { cups.IPP_JOB_PENDING: _("Pending"),
-                          cups.IPP_JOB_HELD: _("Held"),
-                          cups.IPP_JOB_PROCESSING: _("Processing"),
-                          cups.IPP_JOB_STOPPED: _("Stopped"),
-                          cups.IPP_JOB_CANCELED: _("Canceled"),
-                          cups.IPP_JOB_ABORTED: _("Aborted"),
-                          cups.IPP_JOB_COMPLETED: _("Completed") }[s]
+                job_requires_auth = (jstate == cups.IPP_JOB_HELD and
+                                     data.get ('job-hold-until', 'none') ==
+                                     'auth-info-required')
+                if job_requires_auth:
+                    state = _("Held for authentication")
+                else:
+                    state = { cups.IPP_JOB_PENDING: _("Pending"),
+                              cups.IPP_JOB_HELD: _("Held"),
+                              cups.IPP_JOB_PROCESSING: _("Processing"),
+                              cups.IPP_JOB_STOPPED: _("Stopped"),
+                              cups.IPP_JOB_CANCELED: _("Canceled"),
+                              cups.IPP_JOB_ABORTED: _("Aborted"),
+                              cups.IPP_JOB_COMPLETED: _("Completed") }[s]
             except ValueError:
                 pass
             except IndexError:
@@ -491,75 +564,152 @@ class JobViewer (monitor.Watcher):
         store.set_value (iter, 6, state)
 
         # Check whether authentication is required.
-        if (self.trayicon and
-            data['job-state'] == cups.IPP_JOB_HELD and
-            data.get ('job-hold-until', 'none') == 'auth-info-required' and
-            not self.auth_info_dialog):
-            try:
-                cups.require ("1.9.37")
-            except:
-                debugprint ("Authentication required but "
-                            "authenticateJob() not available")
-                return
-
-            # Find out which auth-info is required.
-            try:
-                c = authconn.Connection (self.MainWindow)
+        if self.trayicon:
+            if (job_requires_auth and
+                not self.auth_notifications.has_key (job) and
+                not self.auth_info_dialogs.has_key (job)):
                 try:
-                    uri = data['job-printer-uri']
-                    attributes = c.getPrinterAttributes (uri = uri)
-                except TypeError: # uri keyword introduced in pycups-1.9.32
-                    debugprint ("Fetching printer attributes by name")
-                    attributes = c.getPrinterAttributes (printer)
+                    cups.require ("1.9.37")
+                except:
+                    debugprint ("Authentication required but "
+                                "authenticateJob() not available")
+                    return
+
+                title = _("Authentication Required")
+                text = _("Job requires authentication to proceed.")
+                notification = pynotify.Notification (title, text, 'printer')
+                notification.set_data ('job-id', job)
+                notification.set_urgency (pynotify.URGENCY_NORMAL)
+                notification.set_timeout (pynotify.EXPIRES_NEVER)
+                notification.connect ('closed',
+                                      self.on_auth_notification_closed)
+                notification.add_action ("authenticate", _("Authenticate"),
+                                         self.on_auth_notification_authenticate)
+                self.auth_notifications[job] = notification
+                debugprint ("auth notification opened for job %s" % job)
+                self.set_statusicon_visibility ()
+
+                # In set_statusicon_visibility we process pending
+                # events, so we need to check that we still have a
+                # notification to show.
+                if notification.get_data ('closed') != True:
+                    notification.attach_to_status_icon (self.statusicon)
+                    notification.show ()
+            elif (not job_requires_auth and
+                  self.auth_notifications.has_key (job)):
+                debugprint ("job %s no longer requires auth" % job)
+                self.auth_notifications[job].close ()
+                self.auth_notifications[job].set_data ('closed', True)
+                del self.auth_notifications[job]
+
+    def on_auth_notification_closed (self, notification, reason=None):
+        job = notification.get_data ('job-id')
+        debugprint ("auth notification closed for job %s" % job)
+        self.auth_notifications[job].set_data ('closed', True)
+        del self.auth_notifications[job]
+
+    def on_auth_notification_authenticate (self, notification, action):
+        job = notification.get_data ('job-id')
+        debugprint ("auth notification authenticate for job %s" % job)
+        self.display_auth_info_dialog (job)
+
+    def display_auth_info_dialog (self, job):
+        data = self.jobs[job]
+        # Find out which auth-info is required.
+        try:
+            c = authconn.Connection (self.MainWindow,
+                                     host=self.host,
+                                     port=self.port,
+                                     encryption=self.encryption)
+            try:
+                uri = data['job-printer-uri']
+                attributes = c.getPrinterAttributes (uri = uri)
+            except TypeError: # uri keyword introduced in pycups-1.9.32
+                debugprint ("Fetching printer attributes by name")
+                attributes = c.getPrinterAttributes (printer)
+        except cups.IPPError, (e, m):
+            self.show_IPP_Error (e, m)
+            return
+        except RuntimeError:
+            debugprint ("Failed to connect when fetching printer attrs")
+            return
+
+        try:
+            auth_info_required = attributes['auth-info-required']
+        except KeyError:
+            debugprint ("No auth-info-required attribute; guessing instead")
+            auth_info_required = ['username', 'password']
+
+        if not isinstance (auth_info_required, list):
+            auth_info_required = [auth_info_required]
+
+        if auth_info_required == ['negotiate']:
+            # Try Kerberos authentication.
+            try:
+                debugprint ("Trying Kerberos auth for job %d" % jobid)
+                c.authenticateJob (jobid)
+            except TypeError:
+                # Requires pycups-1.9.39 for optional auth parameter.
+                # Treat this as a normal job error.
+                debugprint ("... need newer pycups for that")
+                return
             except cups.IPPError, (e, m):
                 self.show_IPP_Error (e, m)
                 return
-            except RuntimeError:
-                debugprint ("Failed to connect when fetching printer attrs")
-                return
 
+        dialog = authconn.AuthDialog (auth_info_required=auth_info_required)
+        dialog.set_position (gtk.WIN_POS_CENTER)
+
+        # Pre-fill 'username' field.
+        if 'username' in auth_info_required:
             try:
-                auth_info_required = attributes['auth-info-required']
-            except KeyError:
-                debugprint ("No auth-info-required attribute; guessing instead")
-                auth_info_required = ['username', 'password']
+                auth_info = map (lambda x: '', auth_info_required)
+                username = pwd.getpwuid (os.getuid ())[0]
+                ind = auth_info_required.index ('username')
+                auth_info[ind] = username
+                dialog.set_auth_info (auth_info)
 
-            if not isinstance (auth_info_required, list):
-                auth_info_required = [auth_info_required]
+                index = 0
+                for field in auth_info_required:
+                    if auth_info[index] == '':
+                        # Focus on the first empty field.
+                        dialog.field_grab_focus (field)
+                        break
+                    index += 1
+            except:
+                nonfatalException ()
 
-            if auth_info_required == ['negotiate']:
-                # Try Kerberos authentication.
-                try:
-                    debugprint ("Trying Kerberos auth for job %d" % jobid)
-                    c.authenticateJob (jobid)
-                except TypeError:
-                    # Requires pycups-1.9.39 for optional auth parameter.
-                    # Treat this as a normal job error.
-                    debugprint ("... need newer pycups for that")
-                    return
-                except cups.IPPError, (e, m):
-                    self.show_IPP_Error (e, m)
-                    return
+        dialog.set_prompt (_("Authentication required for "
+                             "printing document `%s' (job %d)") %
+                           (data.get('job-name', _("Unknown")), job))
+        self.auth_info_dialogs[job] = dialog
+        dialog.connect ('response', self.auth_info_dialog_response)
+        dialog.connect ('delete-event', self.auth_info_dialog_delete)
+        dialog.set_data ('job-id', job)
+        dialog.show_all ()
+        dialog.set_keep_above (True)
+        dialog.show_now ()
+        gtk.gdk.keyboard_grab (dialog.window, True)
+        gtk.gdk.pointer_grab (dialog.window, True)
 
-            dialog = authconn.AuthDialog (auth_info_required=auth_info_required)
-            dialog.set_prompt (_("Authentication required for "
-                                 "printing document `%s' (job %d)") %
-                               (data.get('job-name', _("Unknown")), job))
-            self.auth_info_dialog = dialog
-            dialog.connect ('response', self.auth_info_dialog_response)
-            dialog.set_data ('job-id', job)
-            dialog.show_all ()
+    def auth_info_dialog_delete (self, dialog, event):
+        self.auth_info_dialog_response (dialog, gtk.RESPONSE_CANCEL)
 
     def auth_info_dialog_response (self, dialog, response):
-        dialog.hide ()
-        self.auth_info_dialog = None
+        gtk.gdk.pointer_ungrab ()
+        gtk.gdk.keyboard_ungrab ()
+        jobid = dialog.get_data ('job-id')
+        del self.auth_info_dialogs[jobid]
         if response != gtk.RESPONSE_OK:
+            dialog.destroy ()
             return
 
         auth_info = dialog.get_auth_info ()
-        jobid = dialog.get_data ('job-id')
         try:
-            c = authconn.Connection (self.MainWindow)
+            c = authconn.Connection (self.MainWindow,
+                                     host=self.host,
+                                     port=self.port,
+                                     encryption=self.encryption)
             c.authenticateJob (jobid, auth_info)
         except RuntimeError:
             debugprint ("Error connecting to CUPS for authentication")
@@ -567,6 +717,8 @@ class JobViewer (monitor.Watcher):
             self.show_IPP_Error (e, m)
             pass
 
+        dialog.destroy ()
+
     def set_statusicon_visibility (self):
         if not self.trayicon:
             return
@@ -578,7 +730,10 @@ class JobViewer (monitor.Watcher):
             return
 
         open_notifications = len (self.new_printer_notifications.keys ())
-        open_notifications += len (self.state_reason_notifications.keys ())
+        open_notifications += len (self.auth_notifications.keys ())
+        for reason, notification in self.state_reason_notifications.iteritems():
+            if notification.get_data ('closed') != True:
+                open_notifications += 1
         num_jobs = len (self.jobs.keys ())
 
         debugprint ("open notifications: %d" % open_notifications)
@@ -589,6 +744,10 @@ class JobViewer (monitor.Watcher):
                                      open_notifications > 0 or
                                      num_jobs > self.num_jobs_when_hidden)
 
+        # Let the icon show/hide itself before continuing.
+        while self.process_pending_events and gtk.events_pending ():
+            gtk.main_iteration ()
+
     def on_treeview_popup_menu (self, treeview):
         event = gtk.gdk.Event (gtk.gdk.NOTHING)
         self.show_treeview_popup_menu (treeview, event, 0)
@@ -635,48 +794,78 @@ class JobViewer (monitor.Watcher):
 
     def on_job_cancel_activate(self, menuitem):
         try:
-            c = authconn.Connection (self.MainWindow)
+            c = authconn.Connection (self.MainWindow,
+                                     host=self.host,
+                                     port=self.port,
+                                     encryption=self.encryption)
             c.cancelJob (self.jobid)
             del c
         except cups.IPPError, (e, m):
-            self.show_IPP_Error (e, m)
+            if (e != cups.IPP_NOT_POSSIBLE and
+                e != cups.IPP_NOT_FOUND):
+                self.show_IPP_Error (e, m)
+            self.monitor.refresh ()
             return
         except RuntimeError:
             return
 
+        self.monitor.refresh ()
+
     def on_job_hold_activate(self, menuitem):
         try:
-            c = authconn.Connection (self.MainWindow)
+            c = authconn.Connection (self.MainWindow,
+                                     host=self.host,
+                                     port=self.port,
+                                     encryption=self.encryption)
             c.setJobHoldUntil (self.jobid, "indefinite")
             del c
         except cups.IPPError, (e, m):
-            self.show_IPP_Error (e, m)
+            if (e != cups.IPP_NOT_POSSIBLE and
+                e != cups.IPP_NOT_FOUND):
+                self.show_IPP_Error (e, m)
+            self.monitor.refresh ()
             return
         except RuntimeError:
             return
 
+        self.monitor.refresh ()
+
     def on_job_release_activate(self, menuitem):
         try:
-            c = authconn.Connection (self.MainWindow)
+            c = authconn.Connection (self.MainWindow,
+                                     host=self.host,
+                                     port=self.port,
+                                     encryption=self.encryption)
             c.setJobHoldUntil (self.jobid, "no-hold")
             del c
         except cups.IPPError, (e, m):
-            self.show_IPP_Error (e, m)
+            if (e != cups.IPP_NOT_POSSIBLE and
+                e != cups.IPP_NOT_FOUND):
+                self.show_IPP_Error (e, m)
+            self.monitor.refresh ()
             return
         except RuntimeError:
             return
 
+        self.monitor.refresh ()
+
     def on_job_reprint_activate(self, menuitem):
         try:
-            c = authconn.Connection (self.MainWindow)
+            c = authconn.Connection (self.MainWindow,
+                                     host=self.host,
+                                     port=self.port,
+                                     encryption=self.encryption)
             c.restartJob (self.jobid)
             del c
         except cups.IPPError, (e, m):
             self.show_IPP_Error (e, m)
+            self.monitor.refresh ()
             return
         except RuntimeError:
             return
 
+        self.monitor.refresh ()
+
     def on_refresh_activate(self, menuitem):
         self.monitor.refresh ()
 
@@ -830,33 +1019,27 @@ class JobViewer (monitor.Watcher):
         (title, text) = reason.get_description ()
         notification = pynotify.Notification (title, text, 'printer')
         reason.user_notified = True
-        notification.set_data ('printer-state-reason', reason)
         notification.set_urgency (urgency)
         notification.set_timeout (pynotify.EXPIRES_NEVER)
         notification.connect ('closed',
                               self.on_state_reason_notification_closed)
         self.state_reason_notifications[reason.get_tuple ()] = notification
         self.set_statusicon_visibility ()
-        # Let the icon show itself, ready for the notification
-        while gtk.events_pending ():
-            gtk.main_iteration ()
         notification.attach_to_status_icon (self.statusicon)
         notification.show ()
 
-    def on_state_reason_notification_closed (self, notification):
+    def on_state_reason_notification_closed (self, notification, reason=None):
         debugprint ("Notification %s closed" % repr (notification))
-        reason = notification.get_data ('printer-state-reason')
-        tuple = reason.get_tuple ()
-        if self.state_reason_notifications[tuple] == notification:
-            del self.state_reason_notifications[tuple]
-            self.set_statusicon_visibility ()
-            return
-
-        debugprint ("Unable to find closed notification")
+        notification.set_data ('closed', True)
+        self.set_statusicon_visibility ()
+        return
 
     ## monitor.Watcher interface
     def current_printers_and_jobs (self, mon, printers, jobs):
+        monitor.Watcher.current_printers_and_jobs (self, mon, printers, jobs)
+        self.set_process_pending (False)
         self.store.clear ()
+        self.jobs = {}
         self.jobiters = {}
         self.printer_uri_index = PrinterURIIndex (names=printers)
         connection = None
@@ -869,22 +1052,7 @@ class JobViewer (monitor.Watcher):
                 printer = uri
             jobdata['job-printer-name'] = printer
 
-            self.add_job (jobid, jobdata)
-
-            # Fetch complete attributes for these jobs.
-            attrs = None
-            try:
-                if connection == None:
-                    connection = cups.Connection ()
-                attrs = connection.getJobAttributes (jobid)
-            except RuntimeError:
-                pass
-            except AttributeError:
-                pass
-
-            if attrs:
-                jobdata.update (attrs)
-                self.update_job (jobid, jobdata)
+            self.add_job (jobid, jobdata, connection=connection)
 
         self.jobs = jobs
         self.active_jobs = set()
@@ -892,6 +1060,7 @@ class JobViewer (monitor.Watcher):
             if self.job_is_active (jobdata):
                 self.active_jobs.add (jobid)
 
+        self.set_process_pending (True)
         self.update_status ()
 
     def job_added (self, mon, jobid, eventname, event, jobdata):
@@ -934,6 +1103,12 @@ class JobViewer (monitor.Watcher):
         elif jobid in self.active_jobs:
             self.active_jobs.remove (jobid)
 
+        if (jobdata.has_key ('job-hold-until') and
+            not event.has_key ('job-hold-until')):
+            del jobdata['job-hold-until']
+        self.update_job (jobid, jobdata)
+        jobdata = self.jobs[jobid]
+
         # Look out for stopped jobs.
         if (self.trayicon and eventname == 'job-stopped' and
             not jobid in self.stopped_job_prompts):
@@ -941,24 +1116,12 @@ class JobViewer (monitor.Watcher):
             # of some sort, or it might be that the backend requires
             # authentication.  If the latter, the job will be held not
             # stopped, and the job-hold-until attribute will be
-            # 'auth-info-required'.  This will be checked for in
+            # 'auth-info-required'.  This was already checked for in
             # update_job.
-            if jobdata['job-state'] == cups.IPP_JOB_HELD:
-                try:
-                    # Fetch the job-hold-until attribute, as this is
-                    # not provided in the notification attributes.
-                    c = cups.Connection ()
-                    attrs = c.getJobAttributes (jobid)
-                    jobdata.update (attrs)
-                except cups.IPPError:
-                    pass
-                except RuntimeError:
-                    pass
-
             may_be_problem = True
             if (jobdata['job-state'] == cups.IPP_JOB_HELD and
                 jobdata['job-hold-until'] == 'auth-info-required'):
-                # Leave this to update_job to deal with.
+                # update_job already dealt with this.
                 may_be_problem = False
             else:
                 # Other than that, unfortunately the only
@@ -999,6 +1162,7 @@ class JobViewer (monitor.Watcher):
                                                       notify_text)
 
             if may_be_problem:
+                debugprint ("Problem detected")
                 self.toggle_window_display (self.statusicon, force_show=True)
                 dialog = gtk.Dialog (_("Print Error"), self.MainWindow, 0,
                                      (_("_Diagnose"), gtk.RESPONSE_NO,
@@ -1039,8 +1203,6 @@ class JobViewer (monitor.Watcher):
                 self.stopped_job_prompts.add (jobid)
                 dialog.show_all ()
 
-        self.update_job (jobid, jobdata)
-
     def job_removed (self, mon, jobid, eventname, event):
         monitor.Watcher.job_removed (self, mon, jobid, eventname, event)
         if self.jobiters.has_key (jobid):
@@ -1051,6 +1213,11 @@ class JobViewer (monitor.Watcher):
         if jobid in self.active_jobs:
             self.active_jobs.remove (jobid)
 
+        if self.auth_notifications.has_key (jobid):
+            self.auth_notifications[jobid].close ()
+            self.auth_notifications[jobid].set_data ('closed', True)
+            del self.auth_notifications[jobid]
+
         self.update_status ()
 
     def state_reason_added (self, mon, reason):
@@ -1118,7 +1285,10 @@ class JobViewer (monitor.Watcher):
         tuple = reason.get_tuple ()
         try:
             notification = self.state_reason_notifications[tuple]
-            notification.close ()
+            if notification.get_data ('closed') != True:
+                notification.close ()
+            del self.state_reason_notifications[tuple]
+            self.set_statusicon_visibility ()
         except KeyError:
             pass
 
@@ -1164,7 +1334,9 @@ class JobViewer (monitor.Watcher):
             debugprint ("Unexpected now_connected signal")
             return
 
-        notification.close ()
+        if notification.get_data ('closed') != True:
+            notification.close ()
+            notification.set_data ('closed', True)
 
     def printer_event (self, mon, printer, eventname, event):
         monitor.Watcher.printer_event (self, mon, printer, eventname, event)
